#include <clippy/project_list.hpp>
#include <filesystem>
#include <optional>
#include <sstream>
#include <utils/lock_file.hpp>
#include <utils/filesystem.hpp>

#include <fstream>
#include <mutex>
#include <random>
#include <chrono>

#include <jsoncons/json.hpp>
#include <jsoncons_ext/cbor/cbor.hpp>

Config ProjectList::GetNewConfig(const std::filesystem::path& config_directory) {
  std::lock_guard guard(lock_file_);

  LoadWithoutLock();

  std::mt19937 rnd(std::chrono::system_clock::now().time_since_epoch().count());

  auto RandomSymbol = [&rnd]() {
    int n = rnd() % (26 + 26 + 10);
    if (n < 26) {
      return 'a' + n;
    } else if (n < 26 + 26) {
      return 'A' + n - 26;
    } else {
      return '0' + n - 26 - 26;
    }
  };

  auto GenerateFilename = [&rnd, &RandomSymbol]() {
    std::string filename;
    for (size_t i = 0; i < 6; ++i) {
      filename += RandomSymbol();
    }

    return filename;
  };

  auto filename = GenerateFilename();
  while (true) {
    bool exist_config = false;
    for (auto& project : projects_) {
      if (filename == project.configuration_file.filename()) {
        exist_config = true;
        break;
      }
    }
    if (!exist_config) {
      break;
    }

    filename = GenerateFilename();
  }

  auto path_to_config = config_directory / filename;
  projects_.emplace_back(std::filesystem::current_path(), path_to_config);

  SaveConfig();

  return {path_to_config, std::filesystem::current_path()};
}

void ProjectList::Load() {
  std::lock_guard guard(lock_file_);

  LoadWithoutLock();
}

void ProjectList::OldLoadConfig(const std::string& data) {
  std::stringstream in(data);

  Project current;

  while (in >> current.root_project >> current.configuration_file) {
    projects_.push_back(current);
  }
}

void ProjectList::SaveConfig() {
  jsoncons::json result;

  result["version"] = "0.1";
  result["projects"] = std::vector<std::string>();

  for (auto& project : projects_) {
    jsoncons::json current;
    current["path_to_config"] = std::string(project.configuration_file);
    current["path_root_project"] = std::string(project.root_project);

    std::cout << (bool)project.name << std::endl;
    if (project.name && !project.name.value().empty()) {
      current["name"] = project.name.value();
    }

    result["projects"].emplace_back(current);
  }

  std::ofstream out(path_);
  out << result.to_string();
}

void ProjectList::LoadConfig(const jsoncons::json& data) {
  if (data["version"] != "0.1") {
    throw std::logic_error("unsupported version config");
  }

  Project current;
  for (auto& project : data["projects"].array_range()) {
    current.configuration_file = project["path_to_config"].as<std::string>();
    current.root_project = project["path_root_project"].as<std::string>();
    if (project.contains("name")) {
      current.name = project["name"].as<std::string>();
    } else {
      current.name = std::nullopt;
    }

    projects_.emplace_back(current);
  }
}

void ProjectList::LoadWithoutLock() {
  auto data = utils::LoadFile(path_);

  try {
    LoadConfig(jsoncons::json::parse(data));
  } catch (...) {
    std::cout << "I can't read project lists. Try fix it?";

    std::string result;
    std::getline(std::cin, result);

    if (result == "y") {
      OldLoadConfig(data);
      SaveConfig();
    }
  }
}

Project* ProjectList::GetCurrentProject_() {
  auto current_path = std::filesystem::current_path();

  Project* result = nullptr;
  auto UpdateResult = [&result](Project& p) {
    if (!result) {
      result = &p;
    } else {
      if (*result > p) {
        result = &p;
      }
    }
  };

  for (auto& project : projects_) {
    auto prefix_current_path = current_path;

    if (project.root_project == "/") {
      UpdateResult(project);
      continue;
    }

    while (prefix_current_path != "/") {
      if (prefix_current_path == project.root_project) {
        UpdateResult(project);
      }
      prefix_current_path = prefix_current_path.parent_path();
    }
  }

  return result;
}
